"""This module contains the developer interface for video fetching."""

import requests

from .caption import Caption

YOUTUBE_API_KEY = "AIzaSyAO_FJ2SlqU8Q4STEHLGCilw_Y9_11qcW8"


class Video:
    """
    A class to represent a YouTube video and fetch its details using the YouTube API.

    :param videoId: The ID of the YouTube video used to fetch its details.
    :type videoId: str
    :param retries: The number of retry attempts for fetching video details. Defaults to 5.
    :type retries: int
    """

    def __init__(self, videoId: str, retries: int = 5):
        """
        Initialize a Video object with details fetched from the YouTube API.

        :param videoId: The ID of the YouTube video used to fetch its details.
        :type videoId: str
        :param retries: The number of retry attempts for fetching video details. Defaults to 5.
        :type retries: int
        """
        url = "https://www.youtube.com/youtubei/v1/player"
        params = {
            "videoId"       : videoId,
            "key"           : YOUTUBE_API_KEY,
            "contentCheckOk": "True",
            "racyCheckOk"   : "True",
        }
        json_data = {
            "context": {
                "client": {
                    "clientName"   : "MWEB",
                    "clientVersion": "2.20211109.01.00",
                }
            },
            "api_key": YOUTUBE_API_KEY,
        }

        for attempt in range(retries):
            try:
                response = requests.post(url, params=params, json=json_data)
                response.raise_for_status()
                data = response.json()["videoDetails"]

                video_id = data["videoId"]
                channel_id = data["channelId"]
                thumbnails = data["thumbnail"]["thumbnails"]

                self.url = f"https://youtu.be/{video_id}"
                self.title = data["title"]
                self.video_id = video_id
                self.channel = data["author"]
                self.duration = int(data["lengthSeconds"])
                self.views = int(data["viewCount"])
                self.thumbnail = thumbnails[-1]["url"]
                self.thumbnails = thumbnails
                self.channel_id = channel_id
                self.channel_url = f"https://www.youtube.com/channel/{channel_id}"
                self.keywords = data.get("keywords", [])
                self.metadata = response.json()

                return
            except (requests.RequestException, KeyError) as e:
                if attempt < retries - 1:
                    pass
                else:
                    raise e

    def get_captions(self, include_generated: bool = False):
        """
        Retrieve caption tracks from metadata, filtering out auto-generated captions based on the `include_generated` parameter.

        :param include_generated: A boolean flag that determines whether auto-generated captions should be included in the result or not. If set to `True`, all captions, including auto-generated ones, will be returned. If set to `False`, auto-generated captions will be excluded. Defaults to False.
        :type include_generated: bool
        :return: A list of Caption objects.
        :rtype: list
        """
        captions = []
        for track in (
            self.metadata.get("captions", {})
                .get("playerCaptionsTracklistRenderer", {})
                .get("captionTracks", [])
        ):
            is_auto_generated = "auto-generated" in track.get("name", {}).get(
                "runs", [{}]
            )[0].get("text", "")

            if is_auto_generated and not include_generated:
                continue
            captions.append(
                Caption(track["baseUrl"], track["languageCode"], is_auto_generated)
            )

        return captions



class MetaYTError(Exception):
    """Base exception for inheritance."""


class VideoUnavailable(MetaYTError):
    """
    Exception raised when a video is unavailable.

    :param videoId: The ID of the unavailable video.
    :type videoId: str
    """

    def __init__(self, videoId: str):
        """
        Initializes a `VideoUnavailable` exception instance.

        :param videoId: The ID of the unavailable video.
        :type videoId: str
        """
        self.videoId = videoId
        super().__init__(self._error_message)

    @property
    def _error_message(self) -> str:
        """
        Returns the error message for the VideoUnavailable exception.

        :return: The error message explaining why the video is unavailable.
        :rtype: str
        """
        return (
            f"Video ID: `{self.videoId}` is unavailable, it might be caused by various reasons such as:\n"
            "- Invalid Video ID\n"
            "- Video is private\n"
            "- Video is region locked"
        )



from typing import Dict, List


def convert_to_srt(transcript: List[Dict[str, float]]) -> str:
    """
    Convert a transcript to SubRip (SRT) format.

    :param transcript: The transcript to convert.
    :type transcript: list
    :return: The transcript in SubRip (SRT) format.
    :rtype: str
    """
    srt_output = ""

    for index, item in enumerate(transcript):
        start_time = format_time(item["start"])
        end_time = format_time(item["end"])
        text = item["text"]

        srt_output += f"{index + 1}\n"
        srt_output += f"{start_time} --> {end_time}\n"
        srt_output += f"{text}\n\n"

    return srt_output


def format_time(seconds: float) -> str:
    """
    Format the time in 'HH:MM:SS,MS' format for SRT.

    :param seconds: The time in seconds.
    :type seconds: float
    :return: The formatted time.
    :rtype: str
    """

    hours = int(seconds // 3600)
    minutes = int((seconds % 3600) // 60)
    seconds_remaining = seconds % 60
    milliseconds = round((seconds_remaining - int(seconds_remaining)) * 1000)

    return f"{hours:02}:{minutes:02}:{int(seconds_remaining):02},{milliseconds:03}"



import json
import urllib.parse  # Module for URL encoding

import requests


class Query:
    """
    A class to perform YouTube search queries and extract video information.

    :param query: The search query.
    :type query: str
    :param max_results: The maximum number of results to retrieve. Defaults to None.
    :type max_results: int, optional
    """

    def __init__(self, query: str, max_results: int = None):
        """
        Initialize a Query object.

        :param query: The search query.
        :type query: str
        :param max_results: The maximum number of results to retrieve. Defaults to None.
        :type max_results: int, optional
        """
        self.query = query
        self.max_results = max_results
        self.__results = []
        self.__search__()

    def __parse__(self, response: str):
        """
        Parse the YouTube search response and extract video information.

        :param response: The raw HTML response from the YouTube search.
        :type response: str
        """
        start_index = response.index("ytInitialData") + len("ytInitialData") + 3
        end_index = response.index("};", start_index) + 1

        data = json.loads(response[start_index:end_index])

        for contents in data["contents"]["twoColumnSearchResultsRenderer"][
            "primaryContents"
        ]["sectionListRenderer"]["contents"]:
            for video in contents["itemSectionRenderer"]["contents"]:
                if "videoRenderer" in video.keys():
                    try:
                        result = {
                            "title"  : video["videoRenderer"]["title"]["runs"][0][
                                "text"
                            ],
                            "videoId": video["videoRenderer"]["videoId"],
                        }
                        self.__results.append(result)
                    except KeyError:
                        continue

    def __search__(self):
        """Perform a YouTube search and parse the results."""
        encoded_query = urllib.parse.quote_plus(self.query)
        query_url = f"https://youtube.com/results?search_query={encoded_query}"
        response = requests.get(query_url)

        retry_limit = 15
        retry_count = 0

        while "ytInitialData" not in response.text and retry_count < retry_limit:
            response = requests.get(query_url)
            retry_count += 1

        self.__parse__(response.text)

    @property
    def results(self) -> list | None:
        """
        Get the search results.

        :return: A list containing the search results or None if no results are found.
        :rtype: list | None
        """
        return self.__results[
               : self.max_results
               ]



"""This module contains the developer interface for captions fetching."""

from html import unescape

import requests
import xmltodict

from .langauge import get_language
from .srt import convert_to_srt


class Caption:
    """
    A class to represent video captions and process them.

    :param baseUrl: The base URL of the video.
    :type baseUrl: str
    :param language_code: The ISO 639-1 language code of the captions.
    :type language_code: str
    :param auto_generated: A boolean indicating if the captions are auto-generated.
    :type auto_generated: bool
    """

    def __init__(self, baseUrl: str, language_code: str, auto_generated: bool):
        """
        Initialize a Caption object.

        :param baseUrl: The base URL of the video.
        :type baseUrl: str
        :param language_code: The ISO 639-1 language code of the captions.
        :type language_code: str
        :param auto_generated: A boolean indicating if the captions are auto-generated.
        :type auto_generated: bool
        """
        self.url = f"https://youtube.com{baseUrl}"
        self.language_code = language_code
        self.language = get_language(language_code)
        self.auto_generated = auto_generated
        self.__transcript = []

        response = requests.get(self.url)
        self.xml_script = xmltodict.parse(response.content)

        for block in self.xml_script["transcript"]["text"]:
            line = {
                "text"    : unescape(block.get("#text", "")),
                "start"   : float(block["@start"]),
                "end"     : round(float(block["@start"]) + float(block["@dur"]), 3),
                "duration": float(block["@dur"]),
            }
            self.__transcript.append(line)

    def to_srt(self) -> str:
        """
        Convert transcript to SubRip (SRT) format.

        :return: The transcript in SubRip (SRT) format.
        :rtype: str
        """

        return convert_to_srt(self.__transcript)

    @property
    def transcript(self) -> list:
        """
        Get the transcript of the video.

        :return: A list containing dictionaries representing each line of the transcript.
        :rtype: list
        """
        return self.__transcript



# Dictionary mapping language codes to language names
from typing import Optional

language_code = {
    "af"     : "Afrikaans",
    "ar"     : "Arabic",
    "az"     : "Azerbaijani",
    "be"     : "Belarusian",
    "bg"     : "Bulgarian",
    "bn"     : "Bengali",
    "bs"     : "Bosnian",
    "ca"     : "Catalan",
    "cs"     : "Czech",
    "da"     : "Danish",
    "de"     : "German",
    "el"     : "Greek",
    "en"     : "English",
    "es"     : "Spanish",
    "et"     : "Estonian",
    "eu"     : "Basque",
    "fa"     : "Persian",
    "fi"     : "Finnish",
    "fr"     : "French",
    "ga"     : "Irish",
    "gl"     : "Galician",
    "gu"     : "Gujarati",
    "he"     : "Hebrew",
    "hi"     : "Hindi",
    "hr"     : "Croatian",
    "hu"     : "Hungarian",
    "hy"     : "Armenian",
    "id"     : "Indonesian",
    "is"     : "Icelandic",
    "it"     : "Italian",
    "ja"     : "Japanese",
    "ka"     : "Georgian",
    "kk"     : "Kazakh",
    "km"     : "Khmer",
    "kn"     : "Kannada",
    "ko"     : "Korean",
    "ky"     : "Kyrgyz",
    "lo"     : "Lao",
    "lt"     : "Lithuanian",
    "lv"     : "Latvian",
    "mk"     : "Macedonian",
    "ml"     : "Malayalam",
    "mn"     : "Mongolian",
    "mr"     : "Marathi",
    "ms"     : "Malay",
    "my"     : "Burmese",
    "ne"     : "Nepali",
    "nl"     : "Dutch",
    "no"     : "Norwegian",
    "pa"     : "Punjabi",
    "pl"     : "Polish",
    "pt"     : "Portuguese",
    "ro"     : "Romanian",
    "ru"     : "Russian",
    "si"     : "Sinhala",
    "sk"     : "Slovak",
    "sl"     : "Slovenian",
    "sq"     : "Albanian",
    "sr"     : "Serbian",
    "sv"     : "Swedish",
    "sw"     : "Swahili",
    "ta"     : "Tamil",
    "te"     : "Telugu",
    "th"     : "Thai",
    "tr"     : "Turkish",
    "uk"     : "Ukrainian",
    "ur"     : "Urdu",
    "uz"     : "Uzbek",
    "vi"     : "Vietnamese",
    "zh-Hans": "Chinese (Simplified)",
    "zh-Hant": "Chinese (Traditional)",
    "zu"     : "Zulu"
}


def get_language(code: str) -> Optional[str]:
    """
    Retrieve the language name based on the provided language code.

    Args:
        code (str): The language code for which the language name is to be retrieved.

    Returns:
        Optional[str]: The language name corresponding to the provided code, or None if not found.
    """
    return language_code.get(code)



import codecs
import os

from setuptools import find_packages, setup

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = "\\n" + fh.read()

setup(
    name="meta_yt",
    version="0.1.8",
    packages=find_packages(),
    author="Rystal-Team",
    author_email="code@rystal.xyz",
    description="A lightweight Python library for fetching YouTube videos' metadata.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/Rystal-Team/Meta-YT",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
    install_requires=[
        "requests",
        "xmltodict",
    ],
)



"""This module provides the highest level interface in this module, allowing interactions with `Video`, `Caption` and `Playlist`"""

from typing import Optional
# Import necessary modules
from urllib import parse  # Module for URL parsing

from .query import Query  # Importing the Query class for performing YouTube searches
from .video import Video  # Importing the Video class for handling YouTube videos


def check_isPlaylist(url: str) -> bool:
    """
    Check if a given URL corresponds to a YouTube playlist.

    Args:
        url (str): The URL to check.

    Returns:
        bool: True if the URL corresponds to a playlist, False otherwise.
    """
    return ("playlist?" in url) or ("list=" in url)


def get_video_id(url: str) -> Optional[str]:
    """
    Extract the video ID from a YouTube video URL.

    Args:
        url (str): The YouTube video URL.

    Returns:
        Optional[str]: The video ID if found, None otherwise.
    """
    query = parse.urlparse(url)
    if query.hostname == "youtu.be":
        return query.path[1:]
    if query.hostname in {"www.youtube.com", "youtube.com"}:
        if query.path == "/watch":
            p = parse.parse_qs(query.query)
            return p["v"][0]
        if query.path[:7] == "/embed/":
            return query.path.split("/")[2]
        if query.path[:3] == "/v/":
            return query.path.split("/")[2]
    return None


class YouTube:
    def __init__(self, query: str):
        """
        Initialize a YouTube object.

        Args:
            query (str): The search query or video URL.
        """
        self.video = None
        result = get_video_id(query)

        if result is None:
            results = Query(query, max_results=1).results
            self.videoId = results[0]["videoId"]
        else:
            self.videoId = result

        self.video = Video(self.videoId)



"""Meta-YT: A lightweight Python library for fetching YouTube videos' metadata."""

__title__ = "Meta-YT"
__author__ = "Rystal-Team"
__license__ = "MIT"
__version__ = "0.1.2-beta"

from .caption import Caption
from .video import Video
from .youtube import YouTube
